MODULE lbcnfd
   !!======================================================================
   !!                       ***  MODULE  lbcnfd  ***
   !! Ocean        : north fold  boundary conditions
   !!======================================================================
   !! History :  3.2  ! 2009-03  (R. Benshila)  Original code 
   !!            3.5  ! 2013-07  (I. Epicoco, S. Mocavero - CMCC) MPP optimization
   !!            4.0  ! 2017-04  (G. Madec) automatique allocation of array argument (use any 3rd dimension)
   !!----------------------------------------------------------------------

   !!----------------------------------------------------------------------
   !!   lbc_nfd       : generic interface for lbc_nfd_3d and lbc_nfd_2d routines
   !!   lbc_nfd_3d    : lateral boundary condition: North fold treatment for a 3D arrays   (lbc_nfd)
   !!   lbc_nfd_2d    : lateral boundary condition: North fold treatment for a 2D arrays   (lbc_nfd)
   !!   lbc_nfd_nogather       : generic interface for lbc_nfd_nogather_3d and 
   !!                            lbc_nfd_nogather_2d routines (designed for use
   !!                            with ln_nnogather to avoid global width arrays
   !!                            mpi all gather operations)
   !!----------------------------------------------------------------------
   USE dom_oce        ! ocean space and time domain 
   USE in_out_manager ! I/O manager
   USE lib_mpp        ! MPP library
#if ! defined key_mpi_off
   USE MPI
#endif

   IMPLICIT NONE
   PRIVATE

   INTERFACE lbc_nfd   ! called by mpp_nfd, lbc_lnk_pt2pt, lbc_lnk_neicoll
      MODULE PROCEDURE   lbc_nfd_2d_sp, lbc_nfd_ext_2d_sp, lbc_nfd_3d_sp, lbc_nfd_4d_sp
      MODULE PROCEDURE   lbc_nfd_2d_dp, lbc_nfd_ext_2d_dp, lbc_nfd_3d_dp, lbc_nfd_4d_dp
   END INTERFACE

   INTERFACE lbc_nfd_nogather   ! called by mpp_nfd
      MODULE PROCEDURE   lbc_nfd_nogather_2d_sp, lbc_nfd_nogather_3d_sp, lbc_nfd_nogather_4d_sp
      MODULE PROCEDURE   lbc_nfd_nogather_2d_dp, lbc_nfd_nogather_3d_dp, lbc_nfd_nogather_4d_dp
   END INTERFACE

   INTERFACE mpp_nfd
      MODULE PROCEDURE   mpp_nfd_2d_sp, mpp_nfd_3d_sp, mpp_nfd_4d_sp
      MODULE PROCEDURE   mpp_nfd_2d_dp, mpp_nfd_3d_dp, mpp_nfd_4d_dp
   END INTERFACE
   
   PUBLIC   mpp_nfd            ! mpi north fold conditions
   PUBLIC   lbc_nfd            ! north fold conditions
   PUBLIC   lbc_nfd_nogather   ! north fold conditions (no allgather case)

   INTEGER, PUBLIC, PARAMETER            ::   jpmaxngh = 3               !:
   INTEGER, PUBLIC                       ::   nsndto                     !:
   INTEGER, PUBLIC, DIMENSION (jpmaxngh) ::   isendto                    !: processes to which communicate
   INTEGER, PUBLIC                       ::   ijpj

   !!----------------------------------------------------------------------
   !! NEMO/OCE 4.0 , NEMO Consortium (2018)
   !! $Id$
   !! Software governed by the CeCILL license (see ./LICENSE)
   !!----------------------------------------------------------------------
CONTAINS

   !!----------------------------------------------------------------------
   !!                   ***  routine lbc_nfd_[234]d_[sd]p  ***
   !!               ***  routine lbc_nfd_nogather_[234]d_[sd]p  ***
   !!                   ***  routine lbc_nfd_ext_2d_[sd]p  ***
   !!----------------------------------------------------------------------
   !!
   !! ** Purpose :   lateral boundary condition 
   !!                North fold treatment without processor exchanges. 
   !!
   !! ** Method  :   
   !!
   !! ** Action  :   ptab with updated values along the north fold
   !!----------------------------------------------------------------------
   !
   !                       !==  SINGLE PRECISION VERSIONS
   !
#define PRECISION sp
# define DIM_2d
#    include "lbc_nfd_generic.h90"
#    include "lbc_nfd_nogather_generic.h90"
#    include "lbc_nfd_ext_generic.h90"
# undef DIM_2d
# define DIM_3d
#    include "lbc_nfd_generic.h90"
#    include "lbc_nfd_nogather_generic.h90"
# undef DIM_3d
# define DIM_4d
#    include "lbc_nfd_generic.h90"
#    include "lbc_nfd_nogather_generic.h90"
# undef DIM_4d
#undef PRECISION
   !
   !                       !==  DOUBLE PRECISION VERSIONS
   !
#define PRECISION dp
# define DIM_2d
#    include "lbc_nfd_generic.h90"
#    include "lbc_nfd_nogather_generic.h90"
#    include "lbc_nfd_ext_generic.h90"
# undef DIM_2d
# define DIM_3d
#    include "lbc_nfd_generic.h90"
#    include "lbc_nfd_nogather_generic.h90"
# undef DIM_3d
# define DIM_4d
#    include "lbc_nfd_generic.h90"
#    include "lbc_nfd_nogather_generic.h90"
# undef DIM_4d
#undef PRECISION

   !!======================================================================
   !
   !!----------------------------------------------------------------------
   !!                   ***  routine mpp_nfd_(2,3,4)d  ***
   !!
   !!   * Argument : dummy argument use in mpp_nfd_... routines
   !!                ptab      :   pointer of arrays on which the boundary condition is applied
   !!                cd_nat    :   nature of array grid-points
   !!                psgn      :   sign used across the north fold boundary
   !!                kfld      :   optional, number of pt3d arrays
   !!                kfillmode :   optional, method to be use to fill the halos (see jpfill* variables)
   !!                pfillval  :   optional, background value (used with jpfillcopy)
   !!----------------------------------------------------------------------
   !!
   !!   ----   SINGLE PRECISION VERSIONS
   !!
#define PRECISION sp
# define MPI_TYPE MPI_REAL
# define DIM_2d
#    include "mpp_nfd_generic.h90"
# undef DIM_2d
# define DIM_3d
#    include "mpp_nfd_generic.h90"
# undef DIM_3d
# define DIM_4d
#    include "mpp_nfd_generic.h90"
# undef DIM_4d
# undef MPI_TYPE
#undef PRECISION
   !!
   !!   ----   DOUBLE PRECISION VERSIONS
   !!
#define PRECISION dp
# define MPI_TYPE MPI_DOUBLE_PRECISION
# define DIM_2d
#    include "mpp_nfd_generic.h90"
# undef DIM_2d
# define DIM_3d
#    include "mpp_nfd_generic.h90"
# undef DIM_3d
# define DIM_4d
#    include "mpp_nfd_generic.h90"
# undef DIM_4d
# undef MPI_TYPE
#undef PRECISION

   !!======================================================================
END MODULE lbcnfd
